/////////////////////////////////////////////////////////////
// This source code is Copyright (C) David Wilkinson 1996.
// All rights reserved.
//
// For more information,
//
//    Email: davidw@cascade.demon.co.uk
//    URL:   http://www.netlink.co.uk/users/cascade/http/
//
/////////////////////////////////////////////////////////////


// $Id: HTTPRequest.java,v 1.10 1996/08/04 15:22:33 davidw Exp $
//
// $Log: HTTPRequest.java,v $
// Revision 1.10  1996/08/04 15:22:33  davidw
// Added copyright
//
// Revision 1.9  1996/06/23 12:52:07  davidw
// Use HTTPInformation instead of HTTPEnvironment
// Add logging support
//
// Revision 1.8  1996/06/05 18:09:03  davidw
// Cast InputStream to a DataInutStream to make reading lines easier
//
// Revision 1.7  1996/06/03 20:11:42  davidw
// Use MessageHeaders instead of a hashtable
//
// Revision 1.6  1996/05/06 15:37:25  davidw
// Made into package uk.co.demon.cascade.http
//
// Revision 1.5  1996/05/06 10:42:48  davidw
// Added POST support
//
// Revision 1.4  1996/04/28 14:01:18  davidw
// Yield until Method completes. Also made error messages more verbose.
//
// Revision 1.3  1996/04/28 09:59:30  davidw
// Don't set translated path in here
//
// Revision 1.2  1996/04/14 17:57:33  davidw
// Handle HTTPException properly
//
// Revision 1.1  1996/04/13 22:00:32  davidw
// Initial revision
//

package uk.co.demon.cascade.http;
import uk.co.demon.cascade.http.*;

import java.lang.*;
import java.net.*;
import java.io.*;
import java.util.*;

class HTTPRequest extends Thread
{
    private Socket socket;
    private DataInputStream is;
    private OutputStream os;

    private String methodType;
    private String path;
    private String protocol;

    private HTTPMethod method;
    private MessageHeaders headers;
    private HTTPInformation info;

    public HTTPRequest(Socket s)
    {
	socket = s;

	try
	{
	    is = new DataInputStream(socket.getInputStream());
	    os = socket.getOutputStream();
	}
	catch (IOException e)
	{
	    HTTPException h = new HTTPException("IOException retrieving streams from socket: " + e.getMessage());
	    exceptionHandler (h);
	}

	    
	methodType   = "";
	path         = "";
	protocol     = "";
	
	headers = new MessageHeaders();
	info = new HTTPInformation();
    }


    public void run()
    {
	try 
	{
	    process();
	}
	catch (HTTPException e)
	{
	    exceptionHandler(e);
	}
	finally
	{
	    try
	    {
		os.flush();
		os.close();	
		socket.close();
	    }
	    catch (IOException e)
	    {
		HTTPException h = new HTTPException("IOException closing down request: " + e.getMessage());
		exceptionHandler (h);
	    }
	}
    }



    private void process() throws HTTPException
    {
	getHeaders();

	HTTPSystem.debug("Method Type : " + methodType);
	HTTPSystem.debug("Path        : " + path);
	HTTPSystem.debug("Protocol    : " + protocol);
     
	// Check we have a valid protocol - only support
	// HTTP/0.9 & HTTP/1.0

	if (protocol.equalsIgnoreCase("HTTP/0.9")
	    || protocol.equalsIgnoreCase("HTTP/1.0"))
	{
	    // OK
	}
	else
	{
	    invalidRequest("Unknown Protocol " + protocol);
	}

	method = null;

	if (methodType.equals("GET"))
	{
	    method = new HTTPGetMethod();
	}

	if (methodType.equals("POST"))
	{
	    method = new HTTPPostMethod();
	}

	if (methodType.equals("HEAD"))
	{
	    method = new HTTPHeadMethod();
	}

	if (method == null)
	{
	    // We don't support this method
	    String s = "Method " + methodType + " not supported";
	    HTTPException e = new HTTPException(s);
	    e.setStatus(HTTPStatus.NOT_IMPLEMENTED);
	    throw (e);
	}

	method.setInfo(info);
	method.setPath(path);
	method.setHeaders(headers);
	method.setUpInfo(socket);
	method.setProtocol(protocol);
	method.setInputStream(is);
	method.setOutputStream(os);
	method.process();

	while (method.isAlive())
	{
	    yield();
	}

	processLogs();
    }

    private void getHeaders() throws HTTPException
    {	
	String s;
	boolean firstLineFlag = true;
	
	HTTPSystem.verbose("Reading request header");

	try
	{
	    while ((s = is.readLine()) != null &&  !s.equals(""))
	    {
		HTTPSystem.verbose ("Read header: " + s);
		if (firstLineFlag)
		{
		    // Store request string away for logging
		    info.requestString = s;

		    parseRequestLine(s);
		    if (protocol.equalsIgnoreCase("HTTP/0.9"))
		    {
			// No more headers, so just continue
			break;
		    }
		}
		else
		{
		    parseHeaderLine(s);
		}
		firstLineFlag = false;
	    }
	}
	catch (IOException e)
	{
	    HTTPSystem.verbose("IOException reading headers: " 
			       + e.getMessage());
	}
    }




    private void parseRequestLine(String s) throws HTTPException
    {
	int index = s.trim().indexOf(' ');
	String methodString  = "";
	String pathString    = "";
	String protString    = "";
	String s1            = "";

	if (index == -1)
	{
	    invalidRequest(s);
	}
	else
	{
	    methodString = s.substring(0, index);
	    s1 = s.substring(index).trim();
	}
	
	index = s1.indexOf(' ');

	if (index == -1)
	{
	    // No protocol, so must be HTTP/0.9 Simple-Request format
	    protString = "HTTP/0.9";
	    pathString = s1;
	}
	else
	{
	    pathString = s1.substring(0, index);
	    protString = s1.substring(index).trim();
	}
	
	setMethodType(methodString);
	setPath(pathString);
	setProtocol(protString);
    }

    private void invalidRequest(String s) throws HTTPException
    {
	HTTPException e = new HTTPException("Invalid HTTP Request");
	e.setStatus(HTTPStatus.BAD_REQUEST);
	e.setInfo(s);
	throw(e);
    }



    private void parseHeaderLine(String s)
    {
        try 
	{
	    int colonIndex = s.indexOf(':');
	    if (colonIndex > 0)
	    {
		String type = s.substring(0, colonIndex);
		String value = s.substring(colonIndex + 1);
		headers.put(type, value);
	    }
	}
	catch (StringIndexOutOfBoundsException e)
	{
	    // Bad header line, just ignore it
	    HTTPSystem.verbose("Ignoring bad header line: " + s);
	}
    }


    private void setMethodType(String s)
    {
	methodType = s;
    }


    private void setPath(String s)
    {
	if (s.toString().endsWith("/"))
	{
	    s = s + HTTPConfigurationDefaultDocument.getDefaultDocument();
	}

	path = s;
    }

    private void setProtocol(String s)
    {
	protocol = s;
    }



    private void exceptionHandler(HTTPException e)
    {
	int status = e.getStatus();
	info.status = status;

	try
	{
	    String e_info = e.getInfo();
	    String statusString = HTTPStatus.getString(status);

	    if (statusString == null)
	    {
		// unrecognised status
		status = HTTPStatus.INTERNAL_ERROR;
		statusString = HTTPStatus.getString(status);
	    }
	    
	    String statusMessage = new Integer(status).toString() + " " + statusString;

	    HTTPResponse response = new HTTPResponse(info);

	    response.addHeader("Content-Type", "text/html");
	    response.printStatus(os, status);
	    response.printHeaders(os);
	    
	    StringBuffer message = new StringBuffer();

	    message.append("<html>\n<head>\n<title>" + statusMessage 
			   + "</title>\n</head>");

	    message.append("<body>\n<h1>" + statusMessage + "</h1>\n");

	    message.append(e.getMessage() + "\n");

	    if (e_info != null)
	    {
		message.append("<h2>Additional Info</h2>\n<pre>\n" 
			       + e_info + "\n</pre>\n");
	    }

	    message.append("</body>\n</html>\n");

	    
	    HTTPSystem.output(os, message.toString());
	}
	catch (HTTPException newException)
	{
	    // The exception handler threw an exception
	    // Just ignore it for now

	    HTTPSystem.debug(0, "Exception in exception handler: " 
			     + newException.getMessage());
	}

	if (status >= 200 && status <= 400)
	{
	    // Not an error, so process
	    processLogs();
	}
    }

    private void processLogs()
    {
	HTTPConfigurationLog.processLogs(path, info);
    }
}
